/* maf_proper -- restore "proper order" to merged alignsets, i.e., the property
*  that if alignment A and alignment B both contain sequences S and T, then A
*  precedes B relative to S if and only if A precedes B relative to T.  
*  Suppose maf-file has been created by merging two properly ordered maf files,
*  e.g., using one or several calls to multiz with pairwise alignments that
*  connect the two sets of multiple alignments.  Let the sequences in alignments
*  of the first maf-file be A1, A2, .., Am and let those for the second file be
*  B1, B2, .., Bn.  Then the command:
*  	maf_proper maf-file A1 A2 .. Am + B1 B2 .. Bn > outfile
*  (note the separator "+") "deconstructs" alignments (by puttings rows from
*  B1 .. Bn in their own alignment) until the property is satisfied.
*/

#include "lib.h"
#include "maf.h"
#include "mz_scores.h"
#include "maf_split.h"

//#define VERBOSE

struct mafRec {
	struct mafAli *maf;
	struct mafComp **comp;
};

int main(int argc, char **argv) {
	struct mafFile *mf;
	struct mafRec *mr;
	struct mafAli *a, *b, *A;
	struct mafComp *c, *ip, *iq, *jp, *jq, **array;
	char **source;
	int nseq, ntop, nalign, i, j, p, q, seen_plus;

	argv0 = "maf_proper";
	if (argc < 5)
		fatal("args = maf-file species1 .. + speciesN ..");
	
	nseq = argc - 3;
	source = ckalloc(nseq*sizeof(char *));
	ntop = seen_plus = 0;
	for (j = 0, i = 2; i < argc; ++i)
		if (same_string(argv[i], "+")) {
			if (ntop == 0)
				fatal("saw + immediately");
			seen_plus = 1;
		} else {
			if (!seen_plus)
				++ntop;
			if (j >= nseq)
				fatal("bad command args");
			source[j++] = copy_string(argv[i]);
		}
	if (!seen_plus)
		fatal("needs a + in command line");

	mf = mafReadAll(argv[1], 1);
	A = mf->alignments;
	for (nalign = 0, a = A; a != NULL; a = a->next, ++nalign)
		;
	if (argc > 2) {
		mafWriteStart(stdout, argv0);
		init_scores70();
	}

	mr = ckalloc(nalign*sizeof(struct mafRec));
	array = ckalloc(nalign*nseq*sizeof(struct mafComp *));
	for (a = A, i = 0; i < nalign; ++i, a = a->next) {
		mr[i].maf = a;
		mr[i].comp = &(array[i*nseq]);
		for (j = 0; j < nseq; ++j)
			mr[i].comp[j] = NULL;
		for (c = a->components; c != NULL; c = c->next) {
			for (j = 0; j < nseq; ++j)
				if (same_string(c->src, source[j]))
					break;
			if (j == nseq)
				fatalf("Didn't find %s", c->src);
			if (mr[i].comp[j] != NULL)
				fatalf("Collision over %s", c->src);
			mr[i].comp[j] = c;
		}
	}

	for (i = 0; i < nalign; ++i) {
		if ((a = mr[i].maf) == NULL)
			continue;
		for (j = i+1; j < nalign; ++j) {
			if ((b = mr[j].maf) == NULL)
				continue;
			// Do alignnments i and j have conflicting orders
			// relative to some pair of sequences?
			for (p = 0; p < ntop; ++p) {
				if ((ip = mr[i].comp[p]) == NULL ||
				    (jp = mr[j].comp[p]) == NULL)
					continue;
				for (q = ntop; q < nseq; ++q) {
					if ( (iq = mr[i].comp[q]) == NULL ||
					     (jq = mr[j].comp[q]) == NULL)
						continue;
					if (ip->start < jp->start) {
						if (iq->start > jq->start)
							break;
					} else {
						if (iq->start < jq->start)
							break;
					}
				}
				if (q < nseq)
					break;
			}
			if (p < ntop) {
				// found such a pair of sequences
#ifdef VERBOSE
fprintf(stderr, "for %s and %s, conflict between:\n", source[p], source[q]);
mafWrite(stderr, a);
mafWrite(stderr, b);
#endif
				if (a->textSize <= b->textSize) {
					mafWrite(stdout,
					  split(a, source, ntop, nseq));
					mafWrite(stdout, a);
					mr[i].maf = NULL;
					break;
				} else {
					mafWrite(stdout,
					  split(b, source, ntop, nseq));
					mafWrite(stdout, b);
					mr[j].maf = NULL;
				}
			}
		}
	}
	for (i = 0; i < nalign; ++i)
		if (mr[i].maf != NULL)
	       		mafWrite(stdout, mr[i].maf);

	return 0;
}
