// tba.c -- execute commands to run multiz and satellite programs, creating a
// threaded-block alignment

#define VERSION 5

// arg '-' means no execute, '+' means verbose

#include "lib.h"
#define MZ "multiz"
#define MP "maf_project"
#define MR "maf_proper"
#define MD "maf_porder"
#define MO "maf_order"
#define V2 "lav2maf"
#define F2 "maf2tba"
#define Q2 "seq2tba"
#define TBA_PREFIX "_tba_"
#define DESTINATION "tba.out"
#define DEFAULT_MIN_WIDTH "6"

typedef struct name_list_ {
	char *name;
	struct name_list_ *next;
} NameList, *NameListPtr;

static int execute, verbose;
static char mz[500];	// multiz command perhaps with R= and M= options

/* information at each node of a phylogenetic tree */
typedef struct tree_node_ {
	int id, type;
	NameListPtr names;	/* leaf-species names */
} TreeNode, *TreeNodePtr;

// added 9/5/03
void iterate_porder(int id, char *name_list) {
	char cmd1[10000], cmd2[10000], cmd3[1000];
	int iterate = 1;

	sprintf(cmd1, "%s %stba%d %s > %stemp",
	  MD, TBA_PREFIX, id, name_list, TBA_PREFIX);
	sprintf(cmd2, "%s %stemp > %stba%d", MD, TBA_PREFIX, TBA_PREFIX, id);
	sprintf(cmd3, "mv %stemp %stba%d", TBA_PREFIX, TBA_PREFIX, id);

	while (iterate) {
		iterate = 0;
		if (verbose)
			printf("%s\n", cmd1);
		if (execute && system(cmd1) != 0)
			fatalf("command '%s' failed", cmd1);
		if (verbose)
			printf("%s\n", cmd2);
		if (execute && (iterate = system(cmd2)) != 0) {
			if (verbose)
				printf("%s\n", cmd3);
			system(cmd3);
		}
	}
}

// do_cmd -- print and possible execute a command
void do_cmd(const char *fmt, ...) {
	char buf[10000];

	va_list ap;
	va_start(ap, fmt);
	if (verbose) {
		(void)vprintf(fmt, ap);
		(void)putchar('\n');
		(void)fflush(stdout);
	}
	if (execute) {
		(void)vsprintf(buf, fmt, ap);
		if (system(buf) != 0)
			fatalf("command '%s' failed", buf);
	}
	va_end(ap);
}


static char *bz_cmd(char *x, char *y, int nbz, char **bz_file) {
	int i, n;
	char buf[500];

	sprintf(buf, "%s.%s", x, y);
	n = strlen(buf);
	for (i = 0; i < nbz; ++i)
		if (strncmp(bz_file[i], buf, n) == 0)
			return copy_string(bz_file[i]);
	return NULL;
}

/* merge_trees - generate commands to merge the threaded-block alignments for
*  the two offspring of a node in the phylogenetic tree.
*/
static void merge_trees(TreeNodePtr x, TreeNodePtr y, int id, int nbz,
  char **bz_file) {
	NameListPtr n, n1, n2;
	int found_cmd, single_left, single_right;
	char *cmd, *mp = MP, *v2 = V2, *f2 = F2, *q2 = Q2;

	if ((n1 = x->names) == NULL || (n2 = y->names) == NULL)
		fatal("merge_tree: empty tree");
	if (n1->next == NULL && n2->next == NULL) {
		if ((cmd = bz_cmd(n1->name, n2->name, nbz, bz_file)) == NULL)
			fatalf("no alignment found for %s and %s", n1->name,
			  n2->name);
		//do_cmd("echo \"%s\"", cmd);
		do_cmd("%s %s %s %s > %stemp", v2, cmd, n1->name, n2->name,
		  TBA_PREFIX);
		do_cmd("%s %stemp %s %s > %stba%d",
		  f2, TBA_PREFIX, n1->name, n2->name, TBA_PREFIX, id);
		return;
	}

	single_left = (n1->next == NULL);
	single_right = (n2->next == NULL);
	found_cmd = 0;
	for (n1 = x->names; n1 != NULL; n1 = n1->next)
		for (n2 = y->names; n2 != NULL; n2 = n2->next)
		    if ((cmd = bz_cmd(n1->name, n2->name, nbz, bz_file))
		       != NULL) {
			//printf("\n");
			//do_cmd("echo \"%s\"", cmd);
			if (found_cmd == 0 && single_left) {
				/* The left child is a sequence and this is the
				*  first alignment using it.  We build a tba
				*  for the sequence, broken into pieces that
				*  play well with the pairwise alignment.  Note
				*  that we postponed creating the tba for the
				*  left child until this point so we would have
				*  the appropriate pairwise alignment for
				*  making the one-sequence tba.  Also, we need
				*  to project the child's tba only the sequence
				*  of the first alignment, but that's just the
				*  tba itself, so the set of non-projection
				*  alignments ("_tba_A") is empty.
				*/
				do_cmd("rm -f %sA", TBA_PREFIX);
				do_cmd("touch %sA", TBA_PREFIX);
				// Convert pairwise alignment to maf format.
				do_cmd("%s %s %s %s > %stemp",
				  v2, cmd, n1->name, n2->name, TBA_PREFIX);
				// Apply seq2tba to sequence and alignment
				do_cmd("%s %s %stemp > %sB", q2, n1->name,
				  TBA_PREFIX, TBA_PREFIX);
			} else 
				do_cmd("%s %sleft.maf%d %s %sA > %sB",
				  mp, TBA_PREFIX, id, n1->name, TBA_PREFIX,
				  TBA_PREFIX);
			if (found_cmd == 0 && single_right) {
				// Similar to orphan on left (see above).
				do_cmd("rm -f %sC", TBA_PREFIX);
				do_cmd("touch %sC", TBA_PREFIX);
				do_cmd("%s %s %s %s > %stemp",
				  v2, cmd, n1->name, n2->name, TBA_PREFIX);
				// But now we need to flip rows in the pairwise
				// alignment, i.e., project onto n2->name.
				do_cmd("%s %stemp %s /dev/null > %stemp2",
				  mp, TBA_PREFIX, n2->name, TBA_PREFIX);
				do_cmd("%s %s %stemp2 > %sD", q2, n2->name,
				  TBA_PREFIX, TBA_PREFIX);
			} else 
				do_cmd("%s %sright.maf%d %s %sC > %sD",
				  mp, TBA_PREFIX, id, n2->name, TBA_PREFIX,
				  TBA_PREFIX);
			do_cmd("%s %s %s %s > %stemp", v2, cmd, n1->name,
			  n2->name, TBA_PREFIX);
			do_cmd("%s %stemp %sB %sD %sE %sF nohead >> %stba%d",
			  mz, TBA_PREFIX, TBA_PREFIX, TBA_PREFIX, TBA_PREFIX,
			  TBA_PREFIX, TBA_PREFIX, id);
			do_cmd("cat %shead %sA %sE > %sleft.maf%d",
			  TBA_PREFIX, TBA_PREFIX, TBA_PREFIX, TBA_PREFIX, id);
			do_cmd("cat %shead %sC %sF > %sright.maf%d",
			  TBA_PREFIX, TBA_PREFIX, TBA_PREFIX, TBA_PREFIX, id);
			found_cmd = 1;
		    }
	if (!found_cmd) {
		fprintf(stderr, "No alignments connect tree with leaves:\n");
		for (n = x->names; n; n = n->next)
			fprintf(stderr, "  %s", n->name);
		putc('\n', stderr);
		fprintf(stderr, "and tree with leaves:\n");
		for (n = y->names ; n; n = n->next)
			fprintf(stderr, "  %s", n->name);
		putc('\n', stderr);
		exit(1);
	}

}

#define BIG_BUF_SIZE 50000

int main(int argc, char **argv) {
	char *q, buf[500], *p, *mr = MR, *mo = MO, name_list[10000], cmd[500],
	  big_buf[BIG_BUF_SIZE], *bz_file[10000];
	int i, j, top = -1, id = 0, nbz;
	TreeNode tree[1000];
	NameListPtr n;

	sprintf(cmd, "tba.v%d", VERSION);
	argv0 = cmd;

	if (argc < 3)
	       fatalf("args: [+-] [R=?] [M=?] tree-specs blastz-output-files");
	sprintf(big_buf, "# %s", cmd);
	for (i = 1; i < argc; ++i) {
		if (strlen(big_buf) + strlen(argv[1]) + 2 >= BIG_BUF_SIZE)
			fatal("overflow = big_buf");
		strcat(big_buf, " ");
		strcat(big_buf, argv[i]);
	}

	strcpy(mz, MZ);
	strcat(mz, " ");
	verbose = 0;
	execute = 1;
	buf[0] = '\0';
	if (argc > 1 && same_string(argv[1], "-")) {
		strcat(buf, "-");
		execute = 0;
		verbose = 1;
		--argc;
		++argv;
	} else if (argc > 1 && same_string(argv[1], "+")) {
		strcat(buf, "+");
		verbose = 1;
		--argc;
		++argv;
	}

	while (argc > 1 && strchr("RM",argv[1][0]) && argv[1][1] == '=') {
		strcat(mz, argv[1]);
		strcat(mz, " ");
		--argc;
		++argv;
	}
	if (strchr(mz, 'M') == NULL) {
		strcat(mz, "M=");
		strcat(mz, DEFAULT_MIN_WIDTH);
		strcat(mz, " ");
	}
	if (argc == 4 && same_string(argv[2], "-f")) {
		char buf2[500], *s;
		FILE *fp = ckopen(argv[3], "r");

		for (nbz = 0; fgets(buf2, 500, fp) != NULL; ++nbz) {
			if ((s = strchr(buf2, '\n')) != NULL)
				*s = '\0';
			bz_file[nbz] = copy_string(buf2);
		}
		fclose(fp);
	} else {
		nbz = argc - 2;
		for (i = 0; i < nbz; ++i)
			bz_file[i] = argv[i+2];
	}

	do_cmd("echo \"##maf version=1 scoring=multiz\" > %shead",
	  TBA_PREFIX);

	do_cmd("echo \"##maf version=1 scoring=%s\" > %s", cmd, DESTINATION);
	do_cmd("echo \"%s\" >> %s", big_buf, DESTINATION);

	for (q = argv[1]; *q != '\0'; ++q) {
		if (*q == '(') {
			if (++top >= 1000)
				fatal("parse_tree: stack overflow");
			tree[top].type = '(';
		} else if (*q == ')') {
			if (top < 1 || tree[top].type != 0 ||
				       tree[top-1].type != '(') {
				q[1] = '\0';
				fatalf("parse error: %s", argv[1]);
			}
			tree[top-1] = tree[top];
			--top;
		} else if (isalpha(*q)) {
			/* leaf species */
			if (++top >= 1000)
				fatal("parse_tree: stack overflow");
			p = buf;
			while (isalpha(*q) || isdigit(*q) || *q == '_')
				*p++ = *q++;
			--q;
			*p = '\0';
			n = ckalloc(sizeof(NameList));
			n->name = copy_string(buf);
			n->next = NULL;
			tree[top].id = -1;
			tree[top].type = 0;
			tree[top].names = n;
		} else if (*q != ' ')
			fatalf("improper character in tree specification: %c",
				*q);
		if (top > 0 && tree[top-1].type == 0 && tree[top].type == 0) {
			// Process an internal node by merging its children.
			// One optimization is to identify nodes where both
			// children are leaves, to avoid a use of multiz.
			i = tree[top-1].id;
			j = tree[top].id;
			if (i >= 0)
				do_cmd("mv %stba%d %sleft.maf%d",
				  TBA_PREFIX, i, TBA_PREFIX, id);
			if (j >= 0)
				do_cmd("mv %stba%d %sright.maf%d",
				  TBA_PREFIX, j, TBA_PREFIX, id);
			do_cmd("cp %shead %stba%d",
			  TBA_PREFIX, TBA_PREFIX, id);
			merge_trees(&(tree[top-1]), &(tree[top]), id, nbz,
			  bz_file);
			if (i >= 0 || j >= 0) {
				do_cmd("grep -v maf %sleft.maf%d >> %stba%d",
				  TBA_PREFIX, id, TBA_PREFIX, id);
				do_cmd("grep -v maf %sright.maf%d >> %stba%d",
				  TBA_PREFIX, id, TBA_PREFIX, id);
				name_list[0] = '\0';
				for (n = tree[top-1].names; n != NULL;
				  n = n->next) {
					strcat(name_list, " ");
					strcat(name_list, n->name);
				}
				strcat(name_list, " +");
				for (n = tree[top].names; n != NULL;
				  n = n->next) {
					strcat(name_list, " ");
					strcat(name_list, n->name);
				}
				do_cmd("%s %stba%d %s > %stemp", mr, TBA_PREFIX,
				  id, name_list, TBA_PREFIX);
				do_cmd("mv %stemp %stba%d",
				  TBA_PREFIX, TBA_PREFIX, id);
iterate_porder(id, name_list);
/*
				do_cmd("%s %stba%d %s > %stemp", MD, TBA_PREFIX,
				  id, name_list, TBA_PREFIX);
				do_cmd("%s %stemp > %stba%d", MD, TBA_PREFIX,
				  TBA_PREFIX, id);
*/
			}
			/* merge the lists of names */
			n = tree[top-1].names;
			if (n == NULL)
				fatal("empty list of names");
			while (n->next != NULL)
				n = n->next;
			n->next = tree[top].names;
			--top;
			tree[top].id = id++;
		}
	}
	//printf("\n");
	if ((p = strchr(name_list, '+')) != NULL)
		*p = ' ';
	do_cmd("%s %stba%d %s nohead >> %s", mo, TBA_PREFIX, id-1, name_list,
	  DESTINATION);
	do_cmd("rm %s*", TBA_PREFIX);
	if (top > 0)
		fatal("tree specification contains too many '('");
	if (top != 0 || tree[0].type != 0)
		fatal("tree specification is improper");
	return 0;
}
